@@ -8,13 +8,13 @@ GtkWidget *win;
 GtkWidget *xine;
 char *mrl, *argument;
 
-void test_xine_set_mrl(char *path)
+static void test_xine_set_mrl(char *path)
 {
 	mrl=g_strdup(path);
 	bacon_video_widget_open (BACON_VIDEO_WIDGET (xine), mrl, NULL);
 }
 
-void on_eos_event(GtkWidget *widget, gpointer user_data)
+static void on_eos_event(GtkWidget *widget, gpointer user_data)
 {
 	bacon_video_widget_stop (BACON_VIDEO_WIDGET (xine));
 	bacon_video_widget_close(BACON_VIDEO_WIDGET (xine));
@@ -30,8 +30,8 @@ int main(int argc, char *argv[])
 	guint32 height = 500;
 	guint32 width = 500;
 
-	if (argc < 2) {
-		g_warning ("Need at least 2 arguments");
+	if (argc > 2) {
+		g_warning ("Usage: %s <file>", argv[0]);
 		return 1;
 	}
 
@@ -56,8 +56,8 @@ int main(int argc, char *argv[])
 	gtk_widget_show(xine);
 
 	mrl = NULL;
-	test_xine_set_mrl(argv[1]);
-	argument = g_strdup (argv[1]);
+	test_xine_set_mrl(argv[1] ? argv[1] : LOGO_PATH);
+	argument = g_strdup (argv[1] ? argv[1] : LOGO_PATH);
 	bacon_video_widget_play (BACON_VIDEO_WIDGET (xine), NULL);
 	gtk_main();
 
