@@ -2,7 +2,7 @@
    arch-tag: Implementation of Rhythmbox playlist parser
 
    Copyright (C) 2002, 2003 Bastien Nocera
-   Copyright (C) 2003 Colin Walters <walters@rhythmbox.org>
+   Copyright (C) 2003,2004 Colin Walters <walters@rhythmbox.org>
 
    The Gnome Library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public License as
@@ -28,26 +28,25 @@
 #include "totemplparser-marshal.h"
 
 #include <glib.h>
+#include <glib/gi18n.h>
 #include <gtk/gtk.h>
 #include <libxml/tree.h>
 #include <libxml/parser.h>
 #include <glade/glade.h>
 #include <gconf/gconf-client.h>
-#if HAVE_LIBGNOME_DESKTOP
-#include <libgnome/gnome-desktop-item.h>
-#endif
 #include <libgnomevfs/gnome-vfs.h>
 #include <libgnomevfs/gnome-vfs-mime.h>
 #include <libgnomevfs/gnome-vfs-mime-utils.h>
 #include <string.h>
-#include <libgnome/gnome-i18n.h>
 
 #define READ_CHUNK_SIZE 8192
 #define MIME_READ_CHUNK_SIZE 1024
 
-typedef gboolean (*PlaylistCallback) (TotemPlParser *parser, const char *url, gpointer data);
+typedef TotemPlParserResult (*PlaylistCallback) (TotemPlParser *parser, const char *url, gpointer data);
 static gboolean totem_pl_parser_scheme_is_ignored (TotemPlParser *parser, const char *url);
 static gboolean totem_pl_parser_ignore (TotemPlParser *parser, const char *url);
+static TotemPlParserResult totem_pl_parser_parse_internal (TotemPlParser *parser, const char *url);
+static TotemPlParserResult totem_pl_parser_add_asx (TotemPlParser *parser, const char *url, gpointer data);
 
 typedef struct {
 	char *mimetype;
@@ -57,6 +56,8 @@ typedef struct {
 struct TotemPlParserPrivate
 {
 	GList *ignore_schemes;
+	guint recurse_level;
+	gboolean fallback;
 };
 
 /* Signals */
@@ -73,31 +74,7 @@ static void totem_pl_parser_class_init (TotemPlParserClass *class);
 static void totem_pl_parser_init       (TotemPlParser      *parser);
 static void totem_pl_parser_finalize   (GObject *object);
 
-GtkType
-totem_pl_parser_get_type (void)
-{
-	static GtkType totem_pl_parser_type = 0;
-
-	if (!totem_pl_parser_type) {
-		static const GTypeInfo totem_pl_parser_info = {
-			sizeof (TotemPlParserClass),
-			(GBaseInitFunc) NULL,
-			(GBaseFinalizeFunc) NULL,
-			(GClassInitFunc) totem_pl_parser_class_init,
-			(GClassFinalizeFunc) NULL,
-			NULL /* class_data */,
-			sizeof (TotemPlParser),
-			0 /* n_preallocs */,
-			(GInstanceInitFunc) totem_pl_parser_init,
-		};
-
-		totem_pl_parser_type = g_type_register_static (G_TYPE_OBJECT,
-							   "TotemPlParser", &totem_pl_parser_info,
-							   (GTypeFlags)0);
-	}
-
-	return totem_pl_parser_type;
-}
+G_DEFINE_TYPE(TotemPlParser, totem_pl_parser, G_TYPE_OBJECT)
 
 static void
 totem_pl_parser_class_init (TotemPlParserClass *klass)
@@ -152,7 +129,7 @@ my_gnome_vfs_get_mime_type_with_data (const char *uri, gpointer *data)
 	if (result != GNOME_VFS_OK)
 		return NULL;
 
-	/* Read the whole thing. */
+	/* Read the whole thing, up to MIME_READ_CHUNK_SIZE */
 	buffer = NULL;
 	total_bytes_read = 0;
 	do {
@@ -193,6 +170,23 @@ my_gnome_vfs_get_mime_type_with_data (const char *uri, gpointer *data)
 	return mimetype;
 }
 
+static char*
+totem_pl_parser_base_url (const char *url)
+{
+	/* Yay, let's reconstruct the base by hand */
+	GnomeVFSURI *uri, *parent;
+	char *base;
+
+	uri = gnome_vfs_uri_new (url);
+	parent = gnome_vfs_uri_get_parent (uri);
+	base = gnome_vfs_uri_to_string (parent, 0);
+
+	gnome_vfs_uri_unref (uri);
+	gnome_vfs_uri_unref (parent);
+
+	return base;
+}
+
 static gboolean
 write_string (GnomeVFSHandle *handle, const char *buf, GError **error)
 {
@@ -244,9 +238,48 @@ totem_pl_parser_num_entries (TotemPlParser *parser, GtkTreeModel *model,
 	return num_entries - ignored;
 }
 
-gboolean
-totem_pl_parser_write (TotemPlParser *parser, GtkTreeModel *model,
-		   TotemPlParserIterFunc func, const char *output, GError **error)
+static char *
+totem_pl_parser_relative (const char *url, const char *output)
+{
+	char *url_base, *output_base;
+	char *base, *needle;
+
+	base = NULL;
+	url_base = totem_pl_parser_base_url (url);
+	output_base = totem_pl_parser_base_url (output);
+
+	needle = strstr (url_base, output_base);
+	if (needle != NULL)
+	{
+		GnomeVFSURI *uri;
+		char *newurl;
+
+		uri = gnome_vfs_uri_new (url);
+		newurl = gnome_vfs_uri_to_string (uri, 0);
+		if (newurl[strlen (output_base)] == '/') {
+			base = g_strdup (newurl + strlen (output_base) + 1);
+		} else {
+			base = g_strdup (newurl + strlen (output_base));
+		}
+		gnome_vfs_uri_unref (uri);
+		g_free (newurl);
+
+		/* And finally unescape the string */
+		newurl = gnome_vfs_unescape_string (base, NULL);
+		g_free (base);
+		base = newurl;
+	}
+
+	g_free (url_base);
+	g_free (output_base);
+
+	return base;
+}
+
+static gboolean
+totem_pl_parser_write_pls (TotemPlParser *parser, GtkTreeModel *model,
+		TotemPlParserIterFunc func, const char *output,
+		GError **error)
 {
 	GnomeVFSHandle *handle;
 	GnomeVFSResult res;
@@ -270,8 +303,8 @@ totem_pl_parser_write (TotemPlParser *parser, GtkTreeModel *model,
 		g_set_error(error,
 			    TOTEM_PL_PARSER_ERROR,
 			    TOTEM_PL_PARSER_ERROR_VFS_OPEN,
-			    _("Couldn't open parser: %s"),
-			    gnome_vfs_result_to_string (res));
+			    _("Couldn't open file '%s': %s"),
+			    output, gnome_vfs_result_to_string (res));
 		return FALSE;
 	}
 
@@ -285,7 +318,120 @@ totem_pl_parser_write (TotemPlParser *parser, GtkTreeModel *model,
 	success = write_string (handle, buf, error);
 	g_free (buf);
 	if (success == FALSE)
+	{
+		gnome_vfs_close (handle);
 		return FALSE;
+	}
+
+	for (i = 1; i <= num_entries_total; i++) {
+		GtkTreeIter iter;
+		char *path, *url, *title, *relative;
+
+		path = g_strdup_printf ("%d", i - 1);
+		gtk_tree_model_get_iter_from_string (model, &iter, path);
+		g_free (path);
+
+		func (model, &iter, &url, &title);
+
+		if (totem_pl_parser_scheme_is_ignored (parser, url) != FALSE)
+		{
+			g_free (url);
+			g_free (title);
+			continue;
+		}
+
+		relative = totem_pl_parser_relative (url, output);
+		buf = g_strdup_printf ("File%d=%s\n", i,
+				relative ? relative : url);
+		g_free (relative);
+		g_free (url);
+		success = write_string (handle, buf, error);
+		g_free (buf);
+		if (success == FALSE)
+		{
+			gnome_vfs_close (handle);
+			g_free (title);
+			return FALSE;
+		}
+
+		buf = g_strdup_printf ("Title%d=%s\n", i, title);
+		success = write_string (handle, buf, error);
+		g_free (buf);
+		g_free (title);
+		if (success == FALSE)
+		{
+			gnome_vfs_close (handle);
+			return FALSE;
+		}
+	}
+
+	gnome_vfs_close (handle);
+	return TRUE;
+}
+
+static char *
+totem_pl_parser_url_to_dos (const char *url, const char *output)
+{
+	char *retval, *i;
+
+	retval = totem_pl_parser_relative (url, output);
+
+	if (retval == NULL)
+		retval = g_strdup (url);
+
+	/* Don't change URIs, but change smb:// */
+	if (g_str_has_prefix (retval, "smb://") != FALSE)
+	{
+		char *tmp;
+		tmp = g_strdup (retval + strlen ("smb:"));
+		g_free (retval);
+		retval = tmp;
+	}
+
+	if (strstr (retval, "://") != NULL)
+		return retval;
+
+	i = retval;
+	while (*i != '\0')
+	{
+		if (*i == '/')
+			*i = '\\';
+		i++;
+	}
+
+	return retval;
+}
+
+static gboolean
+totem_pl_parser_write_m3u (TotemPlParser *parser, GtkTreeModel *model,
+		TotemPlParserIterFunc func, const char *output,
+		gboolean dos_compatible, GError **error)
+{
+	GnomeVFSHandle *handle;
+	GnomeVFSResult res;
+	int num_entries_total, i;
+	gboolean success;
+	char *buf;
+
+	res = gnome_vfs_open (&handle, output, GNOME_VFS_OPEN_WRITE);
+	if (res == GNOME_VFS_ERROR_NOT_FOUND) {
+		res = gnome_vfs_create (&handle, output,
+				GNOME_VFS_OPEN_WRITE, FALSE,
+				GNOME_VFS_PERM_USER_WRITE
+				| GNOME_VFS_PERM_USER_READ
+				| GNOME_VFS_PERM_GROUP_READ);
+	}
+
+	if (res != GNOME_VFS_OK) {
+		g_set_error(error,
+			    TOTEM_PL_PARSER_ERROR,
+			    TOTEM_PL_PARSER_ERROR_VFS_OPEN,
+			    _("Couldn't open file '%s': %s"),
+			    output, gnome_vfs_result_to_string (res));
+		return FALSE;
+	}
+
+	num_entries_total = gtk_tree_model_iter_n_children (model, NULL);
 
 	for (i = 1; i <= num_entries_total; i++) {
 		GtkTreeIter iter;
@@ -304,28 +450,55 @@ totem_pl_parser_write (TotemPlParser *parser, GtkTreeModel *model,
 			continue;
 		}
 
-		buf = g_strdup_printf ("File%d=%s\n", i, url);
-		success = write_string (handle, buf, error);
-		g_free (buf);
-		g_free (url);
-		if (success == FALSE)
+		if (dos_compatible != FALSE)
 		{
-			g_free (title);
-			return FALSE;
+			char *dos;
+
+			dos = totem_pl_parser_url_to_dos (url, output);
+			buf = g_strdup_printf ("%s\r\n", dos);
+			g_free (dos);
+		} else {
+			char *relative;
+
+			relative = totem_pl_parser_relative (url, output);
+			buf = g_strdup_printf ("%s\n", relative);
+			g_free (relative);
 		}
 
-		buf = g_strdup_printf ("Title%d=%s\n", i, title);
-		success = write_string (handle, buf, error);
-		g_free (buf);
-		g_free (title);
+		success = write_string (handle, url, error);
+
 		if (success == FALSE)
+		{
+			gnome_vfs_close (handle);
 			return FALSE;
+		}
 	}
 
-	gnome_vfs_close (handle);
 	return TRUE;
 }
 
+gboolean
+totem_pl_parser_write (TotemPlParser *parser, GtkTreeModel *model,
+		TotemPlParserIterFunc func, const char *output,
+		TotemPlParserType type, GError **error)
+{
+	switch (type)
+	{
+	case TOTEM_PL_PARSER_PLS:
+		return totem_pl_parser_write_pls (parser, model, func,
+				output, error);
+	case TOTEM_PL_PARSER_M3U:
+	case TOTEM_PL_PARSER_M3U_DOS:
+		return totem_pl_parser_write_m3u (parser, model, func,
+				output, (type == TOTEM_PL_PARSER_M3U_DOS),
+				error);
+	default:
+		g_assert_not_reached ();
+	}
+
+	return FALSE;
+}
+
 static GnomeVFSResult
 my_eel_read_entire_file (const char *uri,
 		int *file_size,
@@ -385,23 +558,6 @@ my_eel_read_entire_file (const char *uri,
 	return GNOME_VFS_OK;
 }
 
-static char*
-totem_pl_parser_base_url (const char *url)
-{
-	/* Yay, let's reconstruct the base by hand */
-	GnomeVFSURI *uri, *parent;
-	char *base;
-
-	uri = gnome_vfs_uri_new (url);
-	parent = gnome_vfs_uri_get_parent (uri);
-	base = gnome_vfs_uri_to_string (parent, 0);
-
-	gnome_vfs_uri_unref (uri);
-	gnome_vfs_uri_unref (parent);
-
-	return base;
-}
-
 static int
 read_ini_line_int (char **lines, const char *key)
 {
@@ -455,7 +611,7 @@ read_ini_line_string (char **lines, const char *key, gboolean dos_mode)
 				retval[len-2] = '\n';
 				retval[len-1] = '\0';
 			}
-			    
+
 			g_strfreev (bits);
 		}
 	}
@@ -487,7 +643,6 @@ totem_pl_parser_finalize (GObject *object)
 static void
 totem_pl_parser_add_one_url (TotemPlParser *parser, const char *url, const char *title)
 {
-	
 	g_signal_emit (G_OBJECT (parser), totem_pl_parser_table_signals[ENTRY],
 		       0, url, title, NULL);
 }
@@ -500,16 +655,16 @@ totem_pl_parser_add_one_url_ext (TotemPlParser *parser, const char *url, const c
 		       0, url, title, genre);
 }
 
-static gboolean
+static TotemPlParserResult
 totem_pl_parser_add_ram (TotemPlParser *parser, const char *url, gpointer data)
 {
-	gboolean retval = FALSE;
+	gboolean retval = TOTEM_PL_PARSER_RESULT_UNHANDLED;
 	char *contents, **lines;
 	int size, i;
 	const char *split_char;
 
 	if (my_eel_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
-		return FALSE;
+		return TOTEM_PL_PARSER_RESULT_ERROR;
 
 	contents = g_realloc (contents, size + 1);
 	contents[size] = '\0';
@@ -527,13 +682,13 @@ totem_pl_parser_add_ram (TotemPlParser *parser, const char *url, gpointer data)
 		if (strcmp (lines[i], "") == 0)
 			continue;
 
-		retval = TRUE;
+		retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
 
 		/* Either it's a URI, or it has a proper path ... */
 		if (strstr(lines[i], "://") != NULL
 				|| lines[i][0] == G_DIR_SEPARATOR) {
 			/* .ram files can contain .smil entries */
-			if (totem_pl_parser_parse (parser, lines[i]) == FALSE)
+			if (totem_pl_parser_parse_internal (parser, lines[i]) != TOTEM_PL_PARSER_RESULT_SUCCESS)
 			{
 				totem_pl_parser_add_one_url (parser,
 						lines[i], NULL);
@@ -548,7 +703,7 @@ totem_pl_parser_add_ram (TotemPlParser *parser, const char *url, gpointer data)
 			base = totem_pl_parser_base_url (url);
 
 			fullpath = g_strdup_printf ("%s/%s", base, lines[i]);
-			if (totem_pl_parser_parse (parser, fullpath) == FALSE)
+			if (totem_pl_parser_parse_internal (parser, fullpath) != TOTEM_PL_PARSER_RESULT_SUCCESS)
 			{
 				totem_pl_parser_add_one_url (parser, fullpath, NULL);
 			}
@@ -567,7 +722,7 @@ totem_pl_parser_get_extinfo_title (gboolean extinfo, char **lines, int i)
 {
 	const char *retval;
 
-	if (extinfo == FALSE)
+	if (extinfo == FALSE || lines == NULL)
 		return NULL;
 
 	if (i == 0)
@@ -583,10 +738,10 @@ totem_pl_parser_get_extinfo_title (gboolean extinfo, char **lines, int i)
 	return retval;
 }
 
-static gboolean
+static TotemPlParserResult
 totem_pl_parser_add_m3u (TotemPlParser *parser, const char *url, gpointer data)
 {
-	gboolean retval = FALSE;
+	gboolean retval = TOTEM_PL_PARSER_RESULT_UNHANDLED;
 	char *contents, **lines;
 	int size, i;
 	const char *split_char;
@@ -614,7 +769,7 @@ totem_pl_parser_add_m3u (TotemPlParser *parser, const char *url, gpointer data)
 		if (lines[i][0] == '\0')
 			continue;
 
-		retval = TRUE;
+		retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
 
 		/* Ignore comments, but mark it if we have extra info */
 		if (lines[i][0] == '#') {
@@ -657,6 +812,7 @@ totem_pl_parser_add_m3u (TotemPlParser *parser, const char *url, gpointer data)
 					totem_pl_parser_get_extinfo_title (extinfo, lines, i));
 			g_free (fullpath);
 			g_free (base);
+			extinfo = FALSE;
 		}
 	}
 
@@ -665,16 +821,16 @@ totem_pl_parser_add_m3u (TotemPlParser *parser, const char *url, gpointer data)
 	return retval;
 }
 
-static gboolean
+static TotemPlParserResult
 totem_pl_parser_add_asf_parser (TotemPlParser *parser, const char *url,
 		gpointer data)
 {
-	gboolean retval = FALSE;
+	gboolean retval = TOTEM_PL_PARSER_RESULT_UNHANDLED;
 	char *contents, **lines, *ref;
 	int size;
 
 	if (my_eel_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
-		return FALSE;
+		return TOTEM_PL_PARSER_RESULT_ERROR;
 
 	contents = g_realloc (contents, size + 1);
 	contents[size] = '\0';
@@ -684,39 +840,40 @@ totem_pl_parser_add_asf_parser (TotemPlParser *parser, const char *url,
 
 	ref = read_ini_line_string (lines, "Ref1", FALSE);
 
-	if (ref == NULL)
-		goto bail;
+	if (ref == NULL) {
+		g_strfreev (lines);
+		return totem_pl_parser_add_asx (parser, url, data);
+	}
 
 	/* change http to mms, thanks Microsoft */
-	if (strncmp ("http", ref, 4) == 0)
+	if (g_str_has_prefix (ref, "http") != FALSE)
 		memcpy(ref, "mmsh", 4);
 
 	totem_pl_parser_add_one_url (parser, ref, NULL);
-	retval = TRUE;
+	retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
 	g_free (ref);
 
-bail:
 	g_strfreev (lines);
 
 	return retval;
 }
 
-static gboolean
+static TotemPlParserResult
 totem_pl_parser_add_pls (TotemPlParser *parser, const char *url, gpointer data)
 {
-	gboolean retval = FALSE;
+	gboolean retval = TOTEM_PL_PARSER_RESULT_UNHANDLED;
 	char *contents, **lines;
 	int size, i, num_entries;
 	char *split_char;
 	gboolean dos_mode = FALSE;
 
 	if (my_eel_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
-		return FALSE;
+		return TOTEM_PL_PARSER_RESULT_ERROR;
 
 	if (size == 0)
 	{
 		g_free (contents);
-		return TRUE;
+		return TOTEM_PL_PARSER_RESULT_SUCCESS;
 	}
 
 	contents = g_realloc (contents, size + 1);
@@ -724,8 +881,9 @@ totem_pl_parser_add_pls (TotemPlParser *parser, const char *url, gpointer data)
 
 	/* figure out whether we're a unix pls or dos pls */
 	if (strstr(contents,"\x0d") == NULL)
+	{
 		split_char = "\n";
-	else {
+	} else {
 		split_char = "\x0d\n";
 		dos_mode = TRUE;
 	}
@@ -742,7 +900,7 @@ totem_pl_parser_add_pls (TotemPlParser *parser, const char *url, gpointer data)
 	if (num_entries == -1)
 		goto bail;
 
-	retval = TRUE;
+	retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
 
 	for (i = 1; i <= num_entries; i++) {
 		char *file, *title, *genre;
@@ -778,7 +936,7 @@ bail:
 
 static gboolean
 parse_asx_entry (TotemPlParser *parser, char *base, xmlDocPtr doc,
-		xmlNodePtr parent)
+		xmlNodePtr parent, const char *pl_title)
 {
 	xmlNodePtr node;
 	char *title, *url;
@@ -810,15 +968,17 @@ parse_asx_entry (TotemPlParser *parser, char *base, xmlDocPtr doc,
 	}
 
 	if (strstr (url, "://") != NULL || url[0] == '/') {
-		totem_pl_parser_add_one_url (parser, url, title);
+		totem_pl_parser_add_one_url (parser, url,
+				title ? title : pl_title);
 		retval = TRUE;
 	} else {
 		char *fullpath;
 
 		fullpath = g_strdup_printf ("%s/%s", base, url);
 		/* .asx files can contain references to other .asx files */
-		if (totem_pl_parser_parse (parser, fullpath) == FALSE)
-			totem_pl_parser_add_one_url (parser, fullpath, NULL);
+		if (totem_pl_parser_parse_internal (parser, fullpath) != TOTEM_PL_PARSER_RESULT_SUCCESS)
+			totem_pl_parser_add_one_url (parser, fullpath,
+					title ? title : pl_title);
 
 		g_free (fullpath);
 	}
@@ -833,6 +993,7 @@ static gboolean
 parse_asx_entries (TotemPlParser *parser, char *base, xmlDocPtr doc,
 		xmlNodePtr parent)
 {
+	char *title = NULL;
 	xmlNodePtr node;
 	gboolean retval = FALSE;
 
@@ -840,30 +1001,36 @@ parse_asx_entries (TotemPlParser *parser, char *base, xmlDocPtr doc,
 		if (node->name == NULL)
 			continue;
 
+		if (g_ascii_strcasecmp (node->name, "title") == 0) {
+			title = xmlNodeListGetString(doc, node->children, 1);
+		}
+
 		if (g_ascii_strcasecmp (node->name, "entry") == 0) {
 			/* Whee found an entry here, find the REF and TITLE */
-			if (parse_asx_entry (parser, base, doc, node) != FALSE)
+			if (parse_asx_entry (parser, base, doc, node, title) != FALSE)
 				retval = TRUE;
 		}
 		if (g_ascii_strcasecmp (node->name, "entryref") == 0) {
 			/* Found an entryref, give the parent instead of the
 			 * children to the parser */
-			if (parse_asx_entry (parser, base, doc, parent) != FALSE)
+			if (parse_asx_entry (parser, base, doc, parent, title) != FALSE)
 				retval = TRUE;
 		}
 	}
 
+	g_free (title);
+
 	return retval;
 }
 
-static gboolean
+static TotemPlParserResult
 totem_pl_parser_add_asx (TotemPlParser *parser, const char *url, gpointer data)
 {
 	xmlDocPtr doc;
 	xmlNodePtr node;
 	char *contents = NULL, *base;
 	int size;
-	gboolean retval = FALSE;
+	gboolean retval = TOTEM_PL_PARSER_RESULT_UNHANDLED;
 
 	if (my_eel_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
 		return FALSE;
@@ -880,29 +1047,29 @@ totem_pl_parser_add_asx (TotemPlParser *parser, const char *url, gpointer data)
 	if(!doc || !doc->children || !doc->children->name) {
 		if (doc != NULL)
 			xmlFreeDoc(doc);
-		return FALSE;
+		return TOTEM_PL_PARSER_RESULT_ERROR;
 	}
 
 	base = totem_pl_parser_base_url (url);
 
 	for (node = doc->children; node != NULL; node = node->next)
 		if (parse_asx_entries (parser, base, doc, node) != FALSE)
-			retval = TRUE;
+			retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
 
 	g_free (base);
 	xmlFreeDoc(doc);
 	return retval;
 }
 
-static gboolean
+static TotemPlParserResult
 totem_pl_parser_add_ra (TotemPlParser *parser, const char *url, gpointer data)
 {
 	if (data == NULL
-			|| (strncmp (data, "http://", strlen ("http://")) != 0
-			&& strncmp (data, "rtsp://", strlen ("rtsp://")) != 0
-			    && strncmp (data, "pnm://", strlen ("pnm://")) != 0)) {
+			|| (g_str_has_prefix (data, "http://") == FALSE
+			&& g_str_has_prefix (data, "rtsp://") == FALSE
+			&& g_str_has_prefix (data, "pnm://") == FALSE)) {
 		totem_pl_parser_add_one_url (parser, url, NULL);
-		return TRUE;
+		return TOTEM_PL_PARSER_RESULT_SUCCESS;
 	}
 
 	return totem_pl_parser_add_ram (parser, url, NULL);
@@ -987,17 +1154,17 @@ parse_smil_entries (TotemPlParser *parser, char *base, xmlDocPtr doc,
 	return retval;
 }
 
-static gboolean
+static TotemPlParserResult
 totem_pl_parser_add_smil (TotemPlParser *parser, const char *url, gpointer data)
 {
 	xmlDocPtr doc;
 	xmlNodePtr node;
 	char *contents = NULL, *base;
 	int size;
-	gboolean retval = FALSE;
+	gboolean retval = TOTEM_PL_PARSER_RESULT_UNHANDLED;
 
 	if (my_eel_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
-		return FALSE;
+		return TOTEM_PL_PARSER_RESULT_ERROR;
 
 	contents = g_realloc (contents, size + 1);
 	contents[size] = '\0';
@@ -1013,96 +1180,114 @@ totem_pl_parser_add_smil (TotemPlParser *parser, const char *url, gpointer data)
 			|| g_ascii_strcasecmp (doc->children->name,
 				"smil") != 0) {
 		if (doc != NULL)
-			xmlFreeDoc(doc);
-		return FALSE;
+			xmlFreeDoc (doc);
+		return TOTEM_PL_PARSER_RESULT_ERROR;
 	}
 
 	base = totem_pl_parser_base_url (url);
 
 	for (node = doc->children; node != NULL; node = node->next)
 		if (parse_smil_entries (parser, base, doc, node) != FALSE)
-			retval = TRUE;
+			retval = TOTEM_PL_PARSER_RESULT_SUCCESS;
 
-	return FALSE;
+	xmlFreeDoc (doc);
+
+	return retval;
 }
 
-static gboolean
+static TotemPlParserResult
 totem_pl_parser_add_asf (TotemPlParser *parser, const char *url, gpointer data)
 {
-	if (data == NULL) {
+	if (data != NULL &&
+		(g_str_has_prefix (data, "[Reference]") == FALSE
+		 && g_ascii_strncasecmp (data, "<ASX", strlen ("<ASX")) != 0)) {
 		totem_pl_parser_add_one_url (parser, url, NULL);
-		return TRUE;
-	}
-
-	if (strncmp (data, "[Reference]", strlen ("[Reference]")) != 0) {
-		totem_pl_parser_add_one_url (parser, url, NULL);
-		return TRUE;
+		return TOTEM_PL_PARSER_RESULT_SUCCESS;
 	}
 
 	return totem_pl_parser_add_asf_parser (parser, url, data);
 }
 
-#if HAVE_LIBGNOME_DESKTOP
-static gboolean
+static TotemPlParserResult
 totem_pl_parser_add_desktop (TotemPlParser *parser, const char *url, gpointer data)
 {
-	GnomeDesktopItem *ditem;
-	int type;
-	gboolean retval;
-	const char *path, *display_name;
+	char *contents, **lines;
+	const char *path, *display_name, *type;
+	int size;
 
-	ditem = gnome_desktop_item_new_from_file (url, 0, NULL);
-	if (ditem == NULL)
-		return FALSE;
+	if (my_eel_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
+		return TOTEM_PL_PARSER_RESULT_ERROR;
 
-	type = gnome_desktop_item_get_entry_type (ditem);
-	if (type != GNOME_DESKTOP_ITEM_TYPE_LINK) {
-		gnome_desktop_item_unref (ditem);
-		return FALSE;
+	contents = g_realloc (contents, size + 1);
+	contents[size] = '\0';
+
+	lines = g_strsplit (contents, "\n", 0);
+	g_free (contents);
+
+	type = read_ini_line_string (lines, "Type", FALSE);
+	if (type == NULL || g_ascii_strcasecmp (type, "Link") != 0) {
+		g_strfreev (lines);
+		return TOTEM_PL_PARSER_RESULT_ERROR;
 	}
 
-	path = gnome_desktop_item_get_string (ditem, "URL");
+	path = read_ini_line_string (lines, "URL", FALSE);
 	if (path == NULL) {
-		gnome_desktop_item_unref (ditem);
-		return FALSE;
+		g_strfreev (lines);
+		return TOTEM_PL_PARSER_RESULT_ERROR;
 	}
 
-	retval = TRUE;
+	display_name = read_ini_line_string (lines, "Name", FALSE);
 
-	display_name = gnome_desktop_item_get_localestring (ditem, "Name");
 	if (totem_pl_parser_ignore (parser, path) == FALSE) {
 		totem_pl_parser_add_one_url (parser, path, display_name);
 	} else {
-		if (totem_pl_parser_parse (parser, path) == FALSE)
+		if (totem_pl_parser_parse_internal (parser, path) != TOTEM_PL_PARSER_RESULT_SUCCESS)
 			totem_pl_parser_add_one_url (parser, path, display_name);
 	}
-	gnome_desktop_item_unref (ditem);
 
-	return retval;
+	g_strfreev (lines);
+
+	return TOTEM_PL_PARSER_RESULT_SUCCESS;
 }
-#endif
 
-static gboolean
+static int
+totem_pl_parser_dir_compare (GnomeVFSFileInfo *a, GnomeVFSFileInfo *b)
+{
+	if (a->name == NULL) {
+		if (b->name == NULL)
+			return 0;
+		else
+			return -1;
+	} else {
+		if (b->name == NULL)
+			return 1;
+		else
+			return strcmp (a->name, b->name);
+	}
+}
+
+static TotemPlParserResult
 totem_pl_parser_add_directory (TotemPlParser *parser, const char *url,
 			   gpointer data)
 {
-	GnomeVFSDirectoryHandle *handle;
-	GnomeVFSFileInfo *info;
+	GList *list, *l;
 	GnomeVFSResult res;
-	gboolean retval = FALSE;
 
-	if (gnome_vfs_directory_open (&handle, url, GNOME_VFS_FILE_INFO_DEFAULT)
-	    != GNOME_VFS_OK)
-		return FALSE;
+	res = gnome_vfs_directory_list_load (&list, url,
+			GNOME_VFS_FILE_INFO_DEFAULT);
+	if (res != GNOME_VFS_OK)
+		return TOTEM_PL_PARSER_RESULT_ERROR;
 
-	info = gnome_vfs_file_info_new ();
-	res = gnome_vfs_directory_read_next (handle, info);
-	while (res == GNOME_VFS_OK) {
+	list = g_list_sort (list, (GCompareFunc) totem_pl_parser_dir_compare);
+	l = list;
+
+	while (l != NULL) {
 		char *str, *fullpath;
+		GnomeVFSFileInfo *info = l->data;
 
 		if (info->name != NULL && (strcmp (info->name, ".") == 0
 					|| strcmp (info->name, "..") == 0)) {
-			res = gnome_vfs_directory_read_next (handle, info);
+			l = l->next;
 			continue;
 		}
 
@@ -1112,17 +1297,17 @@ totem_pl_parser_add_directory (TotemPlParser *parser, const char *url,
 		else
 			fullpath = str;
 
-		if (totem_pl_parser_parse (parser, fullpath) == FALSE)
+		if (totem_pl_parser_parse_internal (parser, fullpath) != TOTEM_PL_PARSER_RESULT_SUCCESS)
 			totem_pl_parser_add_one_url (parser, fullpath, NULL);
 
-		retval = TRUE;
 		g_free (str);
-		res = gnome_vfs_directory_read_next (handle, info);
+		l = l->next;
 	}
 
-	gnome_vfs_directory_close (handle);
-	gnome_vfs_file_info_unref (info);
-	return retval;
+	g_list_foreach (list, (GFunc) gnome_vfs_file_info_unref, NULL);
+	g_list_free (list);
+
+	return TOTEM_PL_PARSER_RESULT_SUCCESS;
 }
 
 /* These ones need a special treatment, mostly parser formats */
@@ -1131,9 +1316,8 @@ static PlaylistTypes special_types[] = {
 	{ "audio/x-ms-asx", totem_pl_parser_add_asx },
 	{ "audio/x-scpls", totem_pl_parser_add_pls },
 	{ "application/x-smil", totem_pl_parser_add_smil },
-#if HAVE_LIBGNOME_DESKTOP
 	{ "application/x-gnome-app-info", totem_pl_parser_add_desktop },
-#endif	
+	{ "application/x-desktop", totem_pl_parser_add_desktop },
 	{ "x-directory/normal", totem_pl_parser_add_directory },
 	{ "video/x-ms-wvx", totem_pl_parser_add_asx },
 	{ "audio/x-ms-wax", totem_pl_parser_add_asx },
@@ -1161,48 +1345,13 @@ totem_pl_parser_scheme_is_ignored (TotemPlParser *parser, const char *url)
 	for (l = parser->priv->ignore_schemes; l != NULL; l = l->next)
 	{
 		const char *scheme = l->data;
-		if (strncmp (url, scheme, strlen (scheme)) == 0)
+		if (g_str_has_prefix (url, scheme) != FALSE)
 			return TRUE;
 	}
 
 	return FALSE;
 }
 
-static gboolean
-totem_pl_parser_add_url_from_data (TotemPlParser *parser, const char *url)
-{
-	const char *mimetype;
-	gboolean retval;
-	gpointer data;
-	guint i;
-
-	mimetype = my_gnome_vfs_get_mime_type_with_data (url, &data);
-	if (mimetype == NULL)
-		return FALSE;
-
-	for (i = 0; i < G_N_ELEMENTS(special_types); i++) {
-		if (mimetype == NULL)
-			break;
-		if (strcmp (special_types[i].mimetype, mimetype) == 0) {
-			retval = (* special_types[i].func) (parser, url, data);
-			g_free (data);
-			return retval;
-		}
-	}
-
-	for (i = 0; i < G_N_ELEMENTS(dual_types); i++) {
-		if (strcmp (dual_types[i].mimetype, mimetype) == 0) {
-			retval = (* dual_types[i].func) (parser, url, data);
-			g_free (data);
-			return retval;
-		}
-	}
-
-	g_free (data);
-
-	return FALSE;
-}
-
 static gboolean
 totem_pl_parser_ignore (TotemPlParser *parser, const char *url)
 {
@@ -1224,36 +1373,79 @@ totem_pl_parser_ignore (TotemPlParser *parser, const char *url)
 		if (strcmp (dual_types[i].mimetype, mimetype) == 0)
 			return FALSE;
 
+	/* It's a remote file that could be an m3u file */
+	if (strcmp (mimetype, "audio/x-mp3") == 0)
+	{
+		if (strstr (url, "m3u") != NULL)
+			return FALSE;
+	}
+
 	return TRUE;
 }
 
-gboolean
-totem_pl_parser_parse (TotemPlParser *parser, const char *url)
+static TotemPlParserResult
+totem_pl_parser_parse_internal (TotemPlParser *parser, const char *url)
 {
 	const char *mimetype;
 	guint i;
+	gpointer data = NULL;
+	gboolean ret = FALSE;
 
-	g_return_val_if_fail (TOTEM_IS_PL_PARSER (parser), FALSE);
-	g_return_val_if_fail (url != NULL, FALSE);
+	if (parser->priv->recurse_level > 4)
+		return TOTEM_PL_PARSER_RESULT_ERROR;
+
+	parser->priv->recurse_level++;
 
 	if (totem_pl_parser_ignore (parser, url) != FALSE)
-		return FALSE;
+		return TOTEM_PL_PARSER_RESULT_UNHANDLED;
 
 	mimetype = gnome_vfs_get_mime_type (url);
+	if (!mimetype)
+		mimetype = my_gnome_vfs_get_mime_type_with_data (url, &data);
 
 	if (mimetype == NULL)
-		return totem_pl_parser_add_url_from_data (parser, url);
+		return TOTEM_PL_PARSER_RESULT_UNHANDLED;
 
-	for (i = 0; i < G_N_ELEMENTS(special_types); i++)
-		if (strcmp (special_types[i].mimetype, mimetype) == 0)
-			return (* special_types[i].func) (parser, url, NULL);
+	for (i = 0; i < G_N_ELEMENTS(special_types); i++) {
+		if (strcmp (special_types[i].mimetype, mimetype) == 0) {
+			ret = (* special_types[i].func) (parser, url, data);
+			g_free (data);
+			break;
+		}
+	}
 
-	for (i = 0; i < G_N_ELEMENTS(dual_types); i++)
-		if (strcmp (dual_types[i].mimetype, mimetype) == 0)
-			return totem_pl_parser_add_url_from_data (parser, url);
+	for (i = 0; i < G_N_ELEMENTS(dual_types); i++) {
+		if (strcmp (dual_types[i].mimetype, mimetype) == 0) {
+			if (data == NULL) {
+				mimetype = my_gnome_vfs_get_mime_type_with_data (url, &data);
+			}
+			ret = (* dual_types[i].func) (parser, url, data);
+			g_free (data);
+			break;
+		}
+	}
 
-	totem_pl_parser_add_one_url (parser, url, NULL);
-	return TRUE;
+	if (ret == FALSE && parser->priv->fallback) {
+		totem_pl_parser_add_one_url (parser, url, NULL);
+		return TOTEM_PL_PARSER_RESULT_SUCCESS;
+	}
+
+	if (ret)
+		return TOTEM_PL_PARSER_RESULT_SUCCESS;
+	else
+		return TOTEM_PL_PARSER_RESULT_UNHANDLED;
+}
+
+TotemPlParserResult
+totem_pl_parser_parse (TotemPlParser *parser, const char *url,
+		       gboolean fallback)
+{
+	g_return_val_if_fail (TOTEM_IS_PL_PARSER (parser), TOTEM_PL_PARSER_RESULT_UNHANDLED);
+	g_return_val_if_fail (url != NULL, TOTEM_PL_PARSER_RESULT_UNHANDLED);
+
+	parser->priv->recurse_level = 0;
+	parser->priv->fallback = fallback;
+	return totem_pl_parser_parse_internal (parser, url);
 }
 
 void
@@ -1266,4 +1458,3 @@ totem_pl_parser_add_ignored_scheme (TotemPlParser *parser,
 		(parser->priv->ignore_schemes, g_strdup (scheme));
 }
 
-
