@@ -24,6 +24,7 @@
 #include "totem-playlist.h"
 
 #include <gtk/gtk.h>
+#include <glib/gi18n.h>
 #include <gdk/gdkkeysyms.h>
 #include <glade/glade.h>
 #include <gconf/gconf-client.h>
@@ -34,23 +35,6 @@
 #include "totem-pl-parser.h"
 #include "debug.h"
 
-#ifdef HAVE_GTK_ONLY
-#ifdef ENABLE_NLS
-#include <libintl.h>
-#define _(String) dgettext(GETTEXT_PACKAGE,String)
-#ifdef gettext_noop
-#define N_(String) gettext_noop(String)
-#else
-#define N_(String) (String)
-#endif /* gettext_noop */
-#else
-#define _(String) (String)
-#define N_(String) (String)
-#endif /* ENABLE_NLS */
-#else
-#include <libgnome/gnome-i18n.h>
-#endif /* HAVE_GTK_ONLY */
-
 #define PL_LEN (gtk_tree_model_iter_n_children (playlist->_priv->model, NULL))
 
 static void ensure_shuffled (TotemPlaylist *playlist, gboolean shuffle);
@@ -129,33 +113,10 @@ static void totem_playlist_class_init (TotemPlaylistClass *class);
 static void totem_playlist_init       (TotemPlaylist      *playlist);
 
 static void init_treeview (GtkWidget *treeview, TotemPlaylist *playlist);
-static gboolean totem_playlist_unset_playing (TotemPlaylist *playlist);
 
-GtkType
-totem_playlist_get_type (void)
-{
-	static GtkType totem_playlist_type = 0;
+#define totem_playlist_unset_playing(x) totem_playlist_set_playing(x, FALSE)
 
-	if (!totem_playlist_type) {
-		static const GTypeInfo totem_playlist_info = {
-			sizeof (TotemPlaylistClass),
-			(GBaseInitFunc) NULL,
-			(GBaseFinalizeFunc) NULL,
-			(GClassInitFunc) totem_playlist_class_init,
-			(GClassFinalizeFunc) NULL,
-			NULL /* class_data */,
-			sizeof (TotemPlaylist),
-			0 /* n_preallocs */,
-			(GInstanceInitFunc) totem_playlist_init,
-		};
-
-		totem_playlist_type = g_type_register_static (GTK_TYPE_DIALOG,
-				"TotemPlaylist", &totem_playlist_info,
-				(GTypeFlags)0);
-	}
-
-	return totem_playlist_type;
-}
+G_DEFINE_TYPE(TotemPlaylist, totem_playlist, GTK_TYPE_DIALOG)
 
 /* Helper functions */
 static gboolean
@@ -266,11 +227,11 @@ totem_playlist_is_media (const char *mrl)
 	if (mrl == NULL)
 		return FALSE;
 
-	if (strncmp ("cdda:", mrl, strlen ("cdda:")) == 0)
+	if (g_str_has_prefix (mrl, "cdda:") != FALSE)
 		return TRUE;
-	if (strncmp ("dvd:",mrl, strlen ("dvd:")) == 0)
+	if (g_str_has_prefix (mrl, "dvd:") != FALSE)
 		return TRUE;
-	if (strncmp ("vcd:", mrl, strlen ("vcd:")) == 0)
+	if (g_str_has_prefix (mrl, "vcd:") != FALSE)
 		return TRUE;
 
 	return FALSE;
@@ -332,7 +293,7 @@ totem_playlist_save_current_playlist (TotemPlaylist *playlist, const char *outpu
 	retval = totem_pl_parser_write (playlist->_priv->parser,
 			playlist->_priv->model,
 			totem_playlist_save_get_iter_func,
-			output, &error);
+			output, TOTEM_PL_PARSER_PLS, &error);
 
 	if (retval == FALSE)
 	{
@@ -574,9 +535,9 @@ selection_changed (GtkTreeSelection *treeselection, TotemPlaylist *playlist)
 	gtk_widget_set_sensitive (down_button, sensitivity);
 }
 
-/* This function checks if the current item is NULL, and try to update it as the
- * first item of the playlist if so. It returns TRUE if there is a current
- * item */
+/* This function checks if the current item is NULL, and try to update it
+ * as the first item of the playlist if so. It returns TRUE if there is a
+ * current item */
 static gboolean
 update_current_from_playlist (TotemPlaylist *playlist)
 {
@@ -610,10 +571,10 @@ totem_playlist_add_files (GtkWidget *widget, TotemPlaylist *playlist)
 	GtkWidget *fs;
 	int response;
 
-	fs = gtk_file_chooser_dialog_new (_("Select files"),
+	fs = gtk_file_chooser_dialog_new (_("Select Movies or Playlists"),
 			GTK_WINDOW (playlist), GTK_FILE_CHOOSER_ACTION_OPEN,
 			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
-			GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
+			GTK_STOCK_ADD, GTK_RESPONSE_ACCEPT,
 			NULL);
 	gtk_dialog_set_default_response (GTK_DIALOG (fs), GTK_RESPONSE_ACCEPT);
 	gtk_file_chooser_set_select_multiple (GTK_FILE_CHOOSER (fs), TRUE);
@@ -813,6 +774,8 @@ totem_playlist_save_files (GtkWidget *widget, TotemPlaylist *playlist)
 	if (response == GTK_RESPONSE_ACCEPT)
 	{
 		char *filename, *tmp;
+		GnomeVFSResult res;
+		GnomeVFSFileInfo info;
 
 		filename = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (fs));
 
@@ -827,7 +790,8 @@ totem_playlist_save_files (GtkWidget *widget, TotemPlaylist *playlist)
 				G_DIR_SEPARATOR_S, NULL);
 		g_free (tmp);
 
-		if (g_file_test (filename, G_FILE_TEST_EXISTS) != FALSE)
+		res = gnome_vfs_get_file_info (filename, &info, GNOME_VFS_FILE_INFO_DEFAULT);
+		if (res != GNOME_VFS_ERROR_NOT_FOUND && res == GNOME_VFS_OK)
 		{
 			GtkWidget *dialog;
 
@@ -989,6 +953,15 @@ totem_playlist_down_files (GtkWidget *widget, TotemPlaylist *playlist)
 static int
 totem_playlist_key_press (GtkWidget *win, GdkEventKey *event, TotemPlaylist *playlist)
 {
+	/* Special case some shortcuts */
+	if (event->state != 0
+			&& (event->state &GDK_CONTROL_MASK))
+	{
+		switch (event->keyval)
+		case GDK_a:
+			gtk_tree_selection_select_all (playlist->_priv->selection);
+			return TRUE;
+	}
 	/* If we have modifiers, and either Ctrl, Mod1 (Alt), or any
 	 * of Mod3 to Mod5 (Mod2 is num-lock...) are pressed, we
 	 * let Gtk+ handle the key */
@@ -1161,7 +1134,7 @@ typedef struct {
 	int random;
 	int index;
 } RandomData;
-                                                                                
+
 static int
 compare_random (gconstpointer ptr_a, gconstpointer ptr_b)
 {
@@ -1441,6 +1414,7 @@ totem_playlist_new (const char *glade_filename, GdkPixbuf *playing_pix)
 		return NULL;
 	}
 
+	gtk_dialog_set_has_separator (GTK_DIALOG (playlist), FALSE);
 	gtk_window_set_title (GTK_WINDOW (playlist), _("Playlist"));
 	gtk_dialog_add_buttons (GTK_DIALOG (playlist),
 			GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
@@ -1562,10 +1536,19 @@ gboolean
 totem_playlist_add_mrl (TotemPlaylist *playlist, const char *mrl,
 		const char *display_name)
 {
+	TotemPlParserResult res;
+
 	g_return_val_if_fail (mrl != NULL, FALSE);
 
-	if (totem_pl_parser_parse (playlist->_priv->parser, mrl) == FALSE)
+	res = totem_pl_parser_parse (playlist->_priv->parser, mrl, TRUE);
+	if (res == TOTEM_PL_PARSER_RESULT_UNHANDLED)
 		return totem_playlist_add_one_mrl (playlist, mrl, display_name);
+	if (res == TOTEM_PL_PARSER_RESULT_ERROR)
+	{
+		totem_playlist_error (_("Playlist error"), _("The playlist '%s' could not be parsed, it might be damaged."), playlist);
+		return FALSE;
+	}
+
 	return TRUE;
 }
 
@@ -1584,6 +1567,94 @@ totem_playlist_clear (TotemPlaylist *playlist)
 	playlist->_priv->current = NULL;
 }
 
+static gboolean
+totem_playlist_current_has_prefix (TotemPlaylist *playlist,
+		const char *prefix)
+{
+	char *current;
+	gboolean retval;
+
+	current = totem_playlist_get_current_mrl (playlist);
+	retval = g_str_has_prefix (current, prefix);
+	g_free (current);
+	return retval;
+}
+
+void
+totem_playlist_clear_with_prefix (TotemPlaylist *playlist, const char *prefix)
+{
+	GList *list = NULL, *l;
+	guint num_items, i;
+
+	num_items = PL_LEN;
+	if (num_items == 0)
+		return;
+
+	if (totem_playlist_current_has_prefix (playlist, prefix) == FALSE)
+	{
+		g_warning ("totem_playlist_clear_with_prefix called while current mrl doesn't have prefix '%s'", prefix);
+		return;
+	}
+
+	for (i = 0; i < num_items; i++)
+	{
+		GtkTreeIter iter;
+		char *index;
+		char *mrl;
+
+		index = g_strdup_printf ("%d", i);
+		if (gtk_tree_model_get_iter_from_string
+				(playlist->_priv->model,
+				 &iter, index) == FALSE)
+		{
+			g_free (index);
+			continue;
+		}
+		g_free (index);
+
+		gtk_tree_model_get (playlist->_priv->model, &iter,
+				URI_COL, &mrl, -1);
+
+		if (g_str_has_prefix (mrl, prefix) != FALSE)
+		{
+			GtkTreePath *path;
+			GtkTreeRowReference *ref;
+
+			path = gtk_tree_path_new_from_indices (i, -1);
+			ref = gtk_tree_row_reference_new
+				(playlist->_priv->model, path);
+			list = g_list_prepend (list, ref);
+			gtk_tree_path_free (path);
+		}
+
+		g_free (mrl);
+	}
+
+
+	for (l = list; l != NULL; l = l->next)
+	{
+		GtkTreePath *path;
+		GtkTreeIter iter;
+
+		path = gtk_tree_row_reference_get_path (l->data);
+		gtk_tree_model_get_iter (playlist->_priv->model, &iter, path);
+		gtk_list_store_remove (GTK_LIST_STORE (playlist->_priv->model),
+				&iter);
+		gtk_tree_path_free (path);
+		gtk_tree_row_reference_free (l->data);
+	}
+	g_list_free (list);
+
+	playlist->_priv->current_shuffled = -1;
+
+	ensure_shuffled (playlist, playlist->_priv->shuffle);
+	gtk_tree_path_free (playlist->_priv->current);
+	playlist->_priv->current = NULL;
+	g_signal_emit (G_OBJECT (playlist),
+			totem_playlist_table_signals[CURRENT_REMOVED],
+			0, NULL);
+}
+
 char
 *totem_playlist_get_current_mrl (TotemPlaylist *playlist)
 {
@@ -1730,8 +1801,7 @@ totem_playlist_set_playing (TotemPlaylist *playlist, gboolean state)
 			&iter,
 			playlist->_priv->current);
 
-	if (&iter == NULL)
-		return FALSE;
+	g_return_val_if_fail (&iter != NULL, FALSE);
 
 	if (state != FALSE)
 		gtk_list_store_set (store, &iter,
@@ -1744,29 +1814,6 @@ totem_playlist_set_playing (TotemPlaylist *playlist, gboolean state)
 	return TRUE;
 }
 
-static gboolean
-totem_playlist_unset_playing (TotemPlaylist *playlist)
-{
-	GtkListStore *store;
-	GtkTreeIter iter;
-
-	if (update_current_from_playlist (playlist) == FALSE)
-		return FALSE;
-
-	store = GTK_LIST_STORE (playlist->_priv->model);
-	gtk_tree_model_get_iter (playlist->_priv->model,
-			&iter,
-			playlist->_priv->current);
-
-	if (&iter == NULL)
-		return FALSE;
-
-	gtk_list_store_set (store, &iter,
-			PIX_COL, NULL,
-			-1);
-	return TRUE;
-}
-
 void
 totem_playlist_set_previous (TotemPlaylist *playlist)
 {
