@@ -21,6 +21,7 @@
 #define __TOTEM_PRIVATE_H__
 
 #include <glade/glade.h>
+#include <glib/gi18n.h>
 #include "totem-remote.h"
 #include "scrsaver.h"
 #include "egg-recent-model.h"
@@ -35,6 +36,12 @@ typedef enum {
 	TOTEM_CONTROLS_FULLSCREEN
 } ControlsVisibility;
 
+typedef enum {
+	STATE_PLAYING,
+	STATE_PAUSED,
+	STATE_STOPPED
+} TotemStates;
+
 struct Totem {
 	/* Control window */
 	GladeXML *xml;
@@ -58,6 +65,7 @@ struct Totem {
 	GtkWidget *volume;
 	GtkAdjustment *voladj;
 	gboolean vol_lock;
+	gboolean vol_fs_lock;
 	gfloat prev_volume;
 	int volume_first_time;
 
@@ -68,10 +76,6 @@ struct Totem {
 	/* exit fullscreen Popup */
 	GtkWidget *exit_popup;
 
-	/* buffering dialog */
-	GtkWidget *buffer_dialog;
-	GtkWidget *buffer_label;
-
 	/* controls management */
 	ControlsVisibility controls_visibility;
 
@@ -81,7 +85,7 @@ struct Totem {
 	GtkAdjustment *fs_seekadj;
 	GtkWidget *fs_volume;
 	GtkAdjustment *fs_voladj;
-	gint control_popup_height;
+	GtkWidget *tcw_time_label;
 
 	guint popup_timeout;
 	gboolean popup_in_progress;
@@ -99,8 +103,8 @@ struct Totem {
 	GConfClient *gc;
 	TotemRemote *remote;
 	BaconMessageConnection *conn;
-	int action;
 	guint32 keypress_time;
+	TotemStates state;
 };
 
 #endif /* __TOTEM_PRIVATE_H__ */
