@@ -35,32 +35,12 @@
 #include <math.h>
 
 #include <glib.h>
-#include <libgnome/gnome-i18n.h>
+#include <glib/gi18n.h>
 
 #include "video-dev.h"
 
 #ifdef __linux__
 
-#ifdef USE_STABLE_LIBGLIB
-static gboolean
-g_str_has_prefix (gchar *haystack, gchar *needle)
-{
-	if (haystack == NULL && needle == NULL) {
-		return TRUE;
-	}
-
-	if (haystack == NULL || needle == NULL) {
-		return FALSE;
-	}
-
-	if (strncmp (haystack, needle, strlen (needle)) == 0) {
-		return TRUE;
-	}
-
-	return FALSE;
-}
-#endif /* USE_STABLE_LIBGLIB */
-
 static char **
 read_lines (char *filename)
 {
